from .cache_strategy import CacheStrategy
from .cache_eviction_policy import DefaultEvictionPolicy
from .lru_eviction_policy import LRUEvictionPolicy
from .ttl_eviction_policy import TTLEvictionPolicy
from .sliding_window_eviction_policy import SlidingWindowEvictionPolicy
from .random_eviction_policy import RandomEvictionPolicy
from .lfu_eviction_policy import LFUEvictionPolicy


from collections import OrderedDict
from typing import Any, Optional, Callable

from . import policies


class Cache:
    cache: OrderedDict
    capacity: int

    def __init__(self, capacity: int, cache_strategy: policies.CacheStrategy = policies.DefaultEvictionPolicy()):
        self.cache: OrderedDict[Any, Any] = OrderedDict()
        self.capacity: int = capacity
        self._cache_strategy = cache_strategy

    def get(self, key: Any) -> Optional[Any]:
        """Retrieve value from cache by provided key - cache hit or miss."""

        if key not in self.cache:
            return None
        self._cache_strategy.on_access(self, key)
        return self.cache.get(key)

    def put(self, key: Any, value: Any) -> None:
        """Insert a key-value pair into the cache."""
        if key in self.cache:
            self._cache_strategy.on_access(self, key)
        else:
            self._cache_strategy.on_insert(self, key)
        self.cache[key] = value

    # region CONVENIENCE METHODS
    def clear(self):
        self.cache.clear()

    def size(self) -> int:
        """ Return the number of items in the cache """
        return len(self.cache)
    # endregion

    # region DECORATOR
    def __call__(self, func: Callable) -> Callable:
        def wrapper(*args, **kwargs):
            # The ars and kwargs will be the key of the cache
            key = args + tuple(kwargs.items())

            # Try to retrieve result from cache
            result = self.get(key)

            # If key not in cache: retrieve and return
            if result is None:
                result = func(*args, **kwargs)
                self.put(key=key, value=result)
            return result

        return wrapper
    # endregion




class LRUCache(Cache):
    def __init__(self, capacity: int):
        super().__init__(capacity=capacity, cache_strategy=policies.LRUEvictionPolicy())


class TTLCache(Cache):
    def __init__(self, capacity: int, ttl_seconds:float):
        super().__init__(capacity=capacity, cache_strategy=policies.TTLEvictionPolicy(ttl_seconds=ttl_seconds))


class SlidingWindowCache(Cache):
    def __init__(self, capacity: int, expiration_seconds:float):
        super().__init__(capacity=capacity, cache_strategy=policies.SlidingWindowEvictionPolicy(expiration_seconds=expiration_seconds))

class LFUCache(Cache):
    def __init__(self, capacity: int, ):
        super().__init__(capacity=capacity, cache_strategy=policies.LFUEvictionPolicy())

class RandomReplaceCache(Cache):
    def __init__(self, capacity: int, ):
        super().__init__(capacity=capacity, cache_strategy=policies.RandomEvictionPolicy())



import time
from collections import OrderedDict
from typing import Any

from .cache_strategy import CacheStrategy


class SlidingWindowEvictionPolicy(CacheStrategy):
    def __init__(self, expiration_seconds: float):
        self.expiration_seconds = expiration_seconds  # Sliding window duration in seconds
        self.timestamps: OrderedDict[Any, float] = OrderedDict()

    def _is_expired(self, key: Any) -> bool:
        """Check if the item is expired based on its timestamp."""
        return (time.time() - self.timestamps[key]) > self.expiration_seconds

    def on_access(self, cache: "Cache", key: Any) -> None:
        """Refresh the item's timestamp to keep it in the sliding window."""
        if self._is_expired(key):
            cache.cache.pop(key, None)
            self.timestamps.pop(key, None)
            return None
        self.timestamps[key] = time.time()
        cache.cache.move_to_end(key=key)
        # return cache.cache.get(key)

    def on_insert(self, cache: "Cache", key: Any) -> None:
        """Insert an item, evicting if necessary based on sliding window."""
        self.timestamps[key] = time.time()
        self.evict(cache=cache)
        # if len(cache.cache) >= cache.capacity:
        #     cache.cache.popitem(last=False)

    def evict(self, cache: "Cache") -> None:
        """Evict items that fall outside the sliding window."""
        current_time = time.time()
        keys_to_evict = [key for key, timestamp in self.timestamps.items()
                         if (current_time - timestamp) > self.expiration_seconds]
        for key in keys_to_evict:
            cache.cache.pop(key, None)
            self.timestamps.pop(key, None)

        if len(cache.cache) >= cache.capacity:
            cache.cache.popitem(last=False)



import time
from collections import OrderedDict
from typing import Any, Dict, DefaultDict, List
from _collections import defaultdict

from .cache_strategy import CacheStrategy


class LFUEvictionPolicy(CacheStrategy):
    def __init__(self):
        self.frequency_dict: DefaultDict[Any, float] = defaultdict(int)
        self.timestamps: OrderedDict[Any, float] = OrderedDict()

    # def _is_expired(self, key: Any) -> bool:
    #     """Check if the item is expired based on its timestamp."""
    #     return (time.time() - self.key_access_count_dict[key]) > self.ttl

    def on_access(self, cache: "Cache", key: Any) -> None:
        """ Move key to the end and update frequency dict if item is accessed """
        self.frequency_dict[key] += 1
        cache.cache.move_to_end(key=key)

    def on_insert(self, cache: "Cache", key: Any) -> None:
        """Insert an item into the cache, evicting if necessary."""
        if len(cache.cache) >= cache.capacity:
            self.evict(cache=cache)
            # cache.cache.popitem(last=False)
        self.timestamps[key] = time.time()
        self.frequency_dict[key] = 0

    def evict(self, cache: "Cache") -> None:
        """Evict items based on expiration and capacity."""

        # Initialize variables to track the least frequent and oldest key
        least_frequency = float('inf')
        oldest_timestamp = float('inf')
        selected_key = None

        for key in self.frequency_dict:
            frequency = self.frequency_dict[key]
            timestamp = self.timestamps[key]

            # Compare based on frequency first, then timestamp if necessary
            if frequency < least_frequency or (frequency == least_frequency and timestamp < oldest_timestamp):
                least_frequency = frequency
                oldest_timestamp = timestamp
                selected_key = key
        cache.cache.pop(selected_key)

        # If the cache is still at capacity: remove the least recently used
        if len(cache.cache) >= cache.capacity:
            cache.cache.popitem(last=False)



from typing import Any

from .cache_strategy import CacheStrategy


class LRUEvictionPolicy(CacheStrategy):
    def on_access(self, cache: "Cache", key: Any) -> Any:
        """Move the accessed item to the end to mark it as most recently used."""
        cache.cache.move_to_end(key=key)

    def on_insert(self, cache: "Cache", key: Any) -> None:
        """ Before item is inserted; check if we need to evict Least Recently Used item """
        if len(cache.cache) >= cache.capacity:
            self.evict(cache=cache)

    def evict(self, cache: "Cache") -> None:
        """Evict the least recently used item."""
        if cache.cache:
            cache.cache.popitem(last=False)



from typing import Any

from .cache_strategy import CacheStrategy


class DefaultEvictionPolicy(CacheStrategy):
    def on_access(self, cache: "Cache", key: Any) -> None:
        """ Method that is called before a key is accessed in the cache."""
        # Does nothing; key order is maintained
        pass

    def on_insert(self, cache: "Cache", key: Any) -> None:
        """Method that is called before a new key-value pair is inserted into the cache."""
        if len(cache.cache) >= cache.capacity:
            self.evict(cache=cache)

    def evict(self, cache: "Cache") -> None:
        """ Remove the oldes ."""
        cache.cache.popitem(last=False)



import time
from collections import OrderedDict
from typing import Any

from .cache_strategy import CacheStrategy


class TTLEvictionPolicy(CacheStrategy):
    def __init__(self, ttl_seconds: float):
        self.ttl = ttl_seconds  # Time-to-live in seconds
        self.timestamps: OrderedDict[Any, float] = OrderedDict()

    def _is_expired(self, key: Any) -> bool:
        """Check if the item is expired based on its timestamp."""
        return (time.time() - self.timestamps[key]) > self.ttl

    def on_access(self, cache: "Cache", key: Any) -> None:
        """Remove the item if it's expired, otherwise update its access time."""
        if self._is_expired(key):
            cache.cache.pop(key, None)
            self.timestamps.pop(key, None)
            return None
        cache.cache.move_to_end(key=key)


    def on_insert(self, cache: "Cache", key: Any) -> None:
        """Insert an item into the cache, evicting if necessary."""
        if len(cache.cache) >= cache.capacity:
            self.evict(cache=cache)
            # cache.cache.popitem(last=False)
        self.timestamps[key] = time.time()

    def evict(self, cache: "Cache") -> None:
        """Evict items based on expiration and capacity."""

        # First remove all expired keys
        expired_keys = [key for key in self.timestamps if self._is_expired(key=key)]
        for key in expired_keys:
            cache.cache.pop(key, None)
            self.timestamps.pop(key, None)

        # If the cache is still at capacity: remove the least recently used
        if len(cache.cache) >= cache.capacity:
            cache.cache.popitem(last=False)



import random
from typing import Any

from .cache_strategy import CacheStrategy


class RandomEvictionPolicy(CacheStrategy):
    def on_access(self, cache: "Cache", key: Any) -> Any:
        """ Do Nothing """
        pass

    def on_insert(self, cache: "Cache", key: Any) -> None:
        """ Before item is inserted; check if we need to evict one """
        if len(cache.cache) >= cache.capacity:
            self.evict(cache=cache)

    def evict(self, cache: "Cache") -> None:
        """Evict the least recently used item."""
        if cache.cache:
            random_key = random.choice(list(cache.cache.keys()))
            cache.cache.pop(random_key)



from abc import ABC, abstractmethod
from typing import Any


class CacheStrategy(ABC):
    @abstractmethod
    def on_access(self, cache: "Cache", key: Any) -> None:
        """ Method that is called before a key is accessed in the cache."""
        pass

    @abstractmethod
    def on_insert(self, cache: "Cache", key: Any) -> None:
        """Method that is called before a new key-value pair is inserted into the cache."""
        pass

    @abstractmethod
    def evict(self, cache: "Cache") -> None:
        """Method that evicts an item based. """
        pass
